projectal API documentation

# projectal

A python client for the Projectal API.

## Getting started

    import projectal
    import os

    # Supply your Projectal server URL and account credentials
    api_base = 'https://yourcompany.projectal.com'
    api_username = os.environ.get('PROJECTAL_USERNAME')
    api_password = os.environ.get('PROJECTAL_PASSWORD')

    # Test communication with server
    status = projectal.status()

    # Test account credentials
    projectal.login()
    details = projectal.auth_details()

## Changelog

### 5.2.0

- Supported handling for new Projectal API rate limiting. When a request is rate limited, will pause for the required waiting period before retrying the request.

### 5.1.0

- Added new Staff function: `projectal.Staff.create_contract()`. Allows creation of a new contract for the given Staff uuId. It's recommended to set a different Department, Position, Start/End Date or Pay Amount for new Contracts to differentiate them.

  Parameters:

  - UUID: uuId of the Source Staff
  - payload: Optional payload to specify updated fields for the new Staff contract
  - end_current_contract (Default=False): Source Staff Contract will have its End Date set to Current Date. Source Staff Start Date must be before Current Date.
  - start_new_contract (Default=False): New Staff Contract will have its Start Date set to Current Date.

- Allow fetching Staff entities with "CONTRACT" link. Will return list of all Contracts for each Staff.

### 5.0.0

- Updated `projectal.login()` to use a basic Dict to store the login cookie instead of an instance of the RequestsCookieJar object from the Python Requests library. This fixes an issue where an indefinite loop can occur when the stored cookie is cleared unexpectedly after attempting to login again after a token expired.
- Update the stored cookie whenever a new cookie is returned by a successful request. This takes advantage of an updated Projectal implementation that refreshes the login cookie periodically. This should avoid the re-authentication procedure in most cases for scripts with long execution times.

### 4.3.2

- Added CompanyType.Division enum, matching new system defaults.

### 4.3.1

- Fixed typo in "Getting started" example.
- Fixed typo in 4.3.0 Changelog.

### 4.3.0

Breaking changes:

- Added "PREDECESSOR_TASK" option when fetching projectal.Task with links
- Predecessor tasks will be returned as a list of tasks under the taskList attribute
- This attribute name is different to what you would see when using the REST API directly (planList). This change was necessary to allow for directly manipulating the list of links and then saving the task entity to commit any changes, since the REST API is expecting a different key for linking calls (taskList).
- Existing Predecessor Task linking methods were also updated to match the new linking functionality. They now work as a reverse linker, automatically inverting the link relationship between the task entities. This more closely matches what you would expect to see based on the Web UI. I.e. When previously calling `some_task.link_predecessor_task(another_task)`, some_task would be set as a predecessor for another_task instead of the other way around. Now another_task would be set as the predecessor for some_task.

### 4.2.2

- `projectal.User.current_user_permissions()` fixed incorrect query.
- `projectal.Webhook.list()` default limit increased to 1000.

### 4.2.1

- Added classes allowing for the management of dynamic enums. The user must have the "List Management" permission to update enums.

  New classes:

  - `projectal.CompanyTypes`
  - `projectal.SkillLevels`
  - `projectal.StaffTypes`
  - `projectal.PriorityLevels`
  - `projectal.ComplexityLevels`
  - `projectal.CurrencyList`

  The current enum can be retrieved with get(), and updated with set(). For example, to return the current SkillLevels enum:

      projectal.SkillLevels.get()

  For each enum the entire list of key value pairs must be provided when calling set(), any existing values that are omitted from the dictionary will be removed, and any additional values will be added.

  To update the SkillLevels enum with a new value:

      new_value_added = {
          "Senior": 10,
          "Mid": 20,
          "Junior": 30,
          # new SkillLevel value "Beginner"
          "Beginner": 40,
      }
      projectal.SkillLevels.set(new_value_added)

  To change the name of a value, set a new key name for the original value:

      updated_value_name = {
          # changing "Senior" SkillLevel to "Expert"
          "Expert": 10,
          "Mid": 20,
          "Junior": 30,
      }
      projectal.SkillLevels.set(updated_value_name)

  To remove an existing value, call set on a dictionary with that value removed:

      value_removed = {
          "Senior": 10,
          "Mid": 20,
          # "Junior" SkillLevel removed
      }
      projectal.SkillLevels.set(new_value_added)

  Updating the CurrencyList works differently to the other enums, since the names of values must match the alphabetic currency code and the value must match the numeric currency code. This will cause an exception if you try to change the name for any values.

  Adding a new currency:

      new_currency_added = {
        "AED": 784
        ...
        # rest of the existing currencies
        ...
        # new currency to add with the alphabetic and numeric code
        "ZWL": 932,
      }
      projectal.CurrencyList.set(new_currency_added)

  Removing an existing currency requires you to provide the numeric code for the currency as a negative value.

      currency_removed = {
        # this currency will be removed
        "AED": -784
        ...
        # rest of the existing currencies
        ...
      }
      projectal.CurrencyList.set(currency_removed)

### 4.2.0

Version 4.2.0 accompanies the release of Projectal 4.1.0

- Minimum Projectal version is now 4.1.0.

- Changed order of applying link types when an entity is initialized, prevents a type error with reverse linking in certain situations.

- `projectal.Task.reset_duration()` now supports adjustments with multi day calendar exceptions.

- `projectal.Task.reset_duration()` location working days override base exceptions.

- `projectal.TaskTemplate.list()` fixed incorrect query when using inherited method.

### 4.1.0

- DateLimit.Max enum value changed from "9999-12-31" to "3000-01-01". This reflects changes to the Projectal backend that defines this as the maximum allowable date value. The front end typically considers this value as equivalent with having no end date.

- Updated requirements.txt version for requests package

- Minimum Projectal version is now 4.0.40

### 4.0.3

- When a dict object is passed to the update class method, it will be converted to the corresponding Entity type. Allows for proper handling of keys that require being treated as links.

### 4.0.2

- Booking entity is now fetched with project field and either staff or resource field.

- Added missing link methods for 'Booking' entity (Note, File)

- Added missing link methods for 'Activity' entity (Booking, Note, File, Rebate)

- Reduced maximum number of link methods to 100 for a single batch request to prevent timeouts under heavy load.

### 4.0.1

- Minimum Projectal version is now 4.0.0.

### 4.0.0

Version 4.0.0 accompanies the release of Projectal 4.0.

- Added the `Activity` entity, new in Projectal 4.0.

- Added the `Booking` entity, new in Projectal 4.0.

### 3.1.1

- Link requests generated by 'projectal.Entity.create()' and 'projectal.Entity.update()' are now executed in batches. This is enabled by default with the 'batch_linking=True' parameter and can be disabled to execute each link request individually. It is recommended to leave this parameter enabled as this can greatly reduce the number of network requests.

### 3.1.0

- Minimum Projectal version is now 3.1.5.

- Added `projectal.Webhook.list_events()`. See API doc for details on how to use.

- Added `deleted_at` parameter to `projectal.Entity.get()`. This value should be a UTC timestamp from a webhook delete event.

- Added `projectal.ldap_sync()` to initiate a user sync with the LDAP/AD service configured in the Projectal server settings.

- Enhanced output of `projectal.Entity.changes()` function when reporting link changes. It no longer dumps the entire before-and-after list with the full content of each linked entity. Now reports three lists: `added`, `updated`, `removed`. Entities within the `updated` list follow the same `old` vs `new` dictionary model for the data attributes within them. E.g:

      resourceList: [
          'added': [],
          'updated': [
              {'uuId': '14eb4c31-0f92-49d1-8b4d-507ab939003e', 'resourceLink': {'utilization': {'old': 0.1, 'new': 0.9}}},
          ],
          'removed': []
      ]

  This should result in slimmer logs that are much easier to understand as the changes are clearly indicated.

### 3.0.2

- Added `projectal.Entity.get_link_definitions()`. Exposes entity link definition dictionary. Consumers can inspect which links an Entity knows about and their internal settings. Link definitions that appear here are the links valid for `links=[]` parameters.

### 3.0.1

- Fixed fetching project with links=['task'] not being available.

- Improved Permission.list(). Now returns a dict with the permission name as key with Permission objects as the value (instead of list of uuIds).

- Added a way to use the aliasing feature of the API (new in Projectal 3.0). Set `api_alias = 'uuid'` to the UUID of a User object and all requests made will be done as that user. Restore this value to None to resume normal operation. (Some rules and limitations apply. See API for more details.)

- Added complete support for the Tags entity (including linkers).

### 3.0

Version 3.0 accompanies the release of Projectal 3.0.

Breaking changes:

- The `links` parameter on `Entity` functions now consumes a list of entity names instead of a comma-separated string. For example:

      # Before:
      projectal.Staff.get('<uuid>', links='skill,location')  # No longer valid
      # Now:
      projectal.Staff.get('<uuid>', links=['skill', 'location'])

- The `projectal.enums.SkillLevel` enum has had all values renamed to match the new values used in Projectal (Junior, Mid, Senior). This includes the properties on Skill entities indicating work time for auto-scheduling (now `juniorLevel`, `midLevel`, `seniorLevel`).

Other changes:

- Working with entity links has changed in this release. The previous methods are still available and continue to work as before, but there is no need to interact with the `projectal.linkers` methods yourself anymore.

  You can now modify the list of links within an entity and save the entity directly. The library will automatically determine how the links have been modified and issue the correct linker methods on your behalf. E.g., you can now do:

      staff = projectal.Staff.get('<uuid>', links=['skill'])
      staff['firstName'] = "New name"  # Field update
      staff['skillList'] = [skill1, skill2, skill3]  # Link update
      staff.save()  # Both changes are saved

      task = projectal.Task.get('<uuid>', links=['stage'])
      task['stage'] = stage1  # Uses a single object instead of list
      task.save()

  See `examples/linking.py` for a more complete demonstration of linking capabilities and limitations.

- Linkers (`projectal.linkers`) can now be given a list of Entities (of one type) to link/unlink/relink in bulk. E.g:

      staff.unlink_skill(skill1)  # Before
      staff.unlink_skill([skill1, skill2, skill3])  # This works now too

- Linkers now strip the payload to only the required fields instead of passing on the entire Entity object. This cuts down on network traffic significantly.

- Linkers now also work in reverse. The Projectal server currently only supports linking entities in one direction (e.g., Company to Staff), which often means writing something like:

      staff.link_location(location)
      company.link_staff(staff)

  The change in direction is not very intuitive and would require you to constantly verify which direction is the one available to you in the documentation.

  Reverse linkers hide this from you and figure out the direction of the relationship for you behind the scenes. So now this is possible, even though the API doesn't strictly support it:

      staff.link_location(location)
      staff.link_company(company)

  Caveat: the documentation for Staff will not list Company links. You will still have to look up the Company documentation for the link description.

- Requesting entity links with the `links=` parameter will now always ensure the link field (e.g., `taskList`) exists in the result, even if there are no links. The server may not always return a value, but we can use a default value ([] for lists, None for dicts).

- Added a `Permission` entity to correctly type Permissions in responses.

- Added a `Tag` entity, new in Projectal 3.0.

- Added `links` parameter to `Company.get_primary_company()`

- `Department.tree()`: now consumes a `holder` Entity object instead of a uuId.

- `Department.tree()`: added `generic_staff` parameter, new in Projectal 3.0.

- Don't break on trailing slash in Projectal URL

- When creating tasks, populate the `projectRef` and `parent` fields in the returned Task object.

- Added convenience functions for matching on fields where you only want one result (e.g match_one()) which return the first match found.

- Update the entity `history()` method for Projectal 3.0. Some new parameters allow you to restrict the history to a particular range or to get only the changes for a webhook timestamp.

- Entity objects can call `.history()` on themselves.

- The library now keeps a reference to the User account that is currently logged in and using the API: `api_auth_details`.

Known issues:

- You cannot save changes to Notes or Calendars via their holding entity. You must save the changes on the Note or Calendar directly. To illustrate:

      staff = projectal.Staff.get(<uuid>, links=['calendar'])
      calendar = staff['calendarList'][0]
      calendar['name'] = 'Calendar 2'

      # Cannot do this - will not pick up the changes
      staff.save()

      # You must do this for now
      calendar.save()

This will be resolved in a future release.

- When creating Notes, the `created` and `modified` values may differ by 1ms in the object you have a reference to compared to what is actually stored in the database.

- Duration calculation is not precise yet (mentioned in 2.1.0)

### 2.1.0

**Breaking changes**: - Getting location calendar is now done on an instance instead of class. So `projectal.Location.calendar(uuid)` is now simply `location.calendar()` - The `CompanyType.Master` enum has been replaced with `CompanyType.Primary`. This was a leftover reference to the Master Company which was renamed in Projectal several versions ago. **Other changes**: - Date conversion functions return None when given None or empty string - Added `Task.reset_duration()` as a basic duration calculator for tasks. This is a work-in-progress and will be gradually improved. The duration calculator takes into consideration the location to remove non-work days from the estimate of working duration. It currently does not work for the time component or `isWorking=True` exceptions. - Change detection in `Entity.changes()` now excludes cases where the server has no value and the new value is None. Saving this change has no effect and would always detect a change until a non-None value is set, which is noisy and generates more network activity.

### 2.0.3

- Better support for calendars.
  - Distinguish between calendar containers ("Calendar") and the calendar items within them ("CalendarItem").
  - Allow CalendarItems to be saved directly. E.G item.save()

- Fix 'holder' parameter in contact/staff/location/task_template not permitting object type. Now consumes uuId or object to match rest of the library.
- `Entity.changes()` has been extended with an `old=True` flag. When this flag is true, the set of changes will now return both the original and the new values. E.g.

md5-ee1d6487305623b06306ece70105f9aa

- Fixed entity link cache causing errors when deleting a link from an entity which has not been fetched with links (deleting from empty list).

### 2.0.2

- Fixed updating Webhook entities

### 2.0.1

- Fixed application ID not being used correctly.

### 2.0.0

- Version 2.0 accompanies the release of Projectal 2.0. There are no major changes since the previous release.
- Expose `Entity.changes()` function. It returns a list of fields on an entity that have changed since fetching it. These are the changes that will be sent over to the server when an update request is made.
- Added missing 'packaging' dependency to requirements.

### 1.2.0

**Breaking changes**:

- Renamed `request_timestamp` to `response_timestamp` to better reflect its purpose.

- Automatic timestamp conversion into dates (introduced in `1.1.0`) has been reverted. All date fields returned from the server remain as UTC timestamps.

  The reason is that date fields on tasks contain a time component and converting them into date strings was erasing the time, resulting in a value that does not match the database.

  Note: the server supports setting date fields using a date string like `2022-04-05`. You may use this if you prefer but the server will always return a timestamp.

  Note: we provide utility functions for easily converting dates from/to timestamps expected by the Projectal server. See: `projectal.date_from_timestamp()`,`projectal.timestamp_from_date()`, and `projectal.timestamp_from_datetime()`.

**Other changes**: - Implement request chunking - for methods that consume a list of entities, we now automatically batch them up into multiple requests to prevent timeouts on really large request. Values are configurable through `chunk_size_read` and `chunk_size_write`. Default values: Read: 1000 items. Write: 200 items. - Added profile get/set functions on entities for easier use. Now you only need to supply the key and the data. E.g: md5-8d7e61a6e1839afb267c1542c84f1afa

- Entity link methods now automatically update the entity's cached list of links. E.g: a task fetched with staff links will have `task['staffList'] = [Staff1,Staff2]`. Before, doing a `task.link_staff(staff)` did not modify the list to reflect the addition. Now, it will turn into `[Staff1,Staff2,Staff3]`. The same applies for update and delete.

  This allows you to modify links and continue working with that object without having to fetch it again to obtain the most recent link data. Be aware that if you acquire the object without requesting the link data as well (e.g: `projectal.Task.get(id, links='STAFF')`), these lists will not accurately reflect what's in the database, only the changes made while the object is held.

- Support new `applicationId` property on login. Set with: `api_application_id`. The application ID is sent back to you in webhooks so you know which application was the source of the event (and you can choose to filter them accordingly).

- Added `Entity.set_readonly()` to allow setting values on entities that will not be sent over to the server when updating/saving the entity.

  The main use case for this is to populate cached entities which you have just created with values you already know about. This is mainly a workaround for the limitation of the server not sending the full object back after creating it, resulting in the client needing to fetch the object in full again if it needs some of the fields set by the server after creation.

  Additionally, some read-only fields will generate an error on the server if included in the update request. This method lets you set these values on newly created objects without triggering this error.

  A common example is setting the `projectRef` of a task you just created.

### 1.1.1

- Add support for 'profiles' API. Profiles are a type of key-value storage that target any entity. Not currently documented.
- Fix handling error message parsing in ProjectalException for batch create operation
- Add `Task.update_order()` to set task order
- Return empty list when GETing empty list instead of failing (no request to server)
- Expose the timestamp returned by requests that modify the database. Use `projectal.request_timestamp` to get the value of the most recent request (None if no timestamp in response)

### 1.1.0

- Minimum Projectal version is now 1.9.4.

Breaking changes:

- Entity `list()` now returns a list of UUIDs instead of full objects. You may provide an `expand` parameter to restore the previous behavior: `Entity.list(expand=True)`. This change is made for performance reasons where you may have thousands of tasks and getting them all may time out. For those cases, we suggest writing a query to filter down to only the tasks and fields you need.
- `Company.get_master_company()` has been renamed to `Company.get_primary_company()` to match the server.
- The following date fields are converted into date strings upon fetch: `startTime`, `closeTime`, `scheduleStart`, `scheduleFinish`. These fields are added or updated using date strings (like `2022-03-02`), but the server returns timestamps (e.g: 1646006400000) upon fetch, which is confusing. This change ensures they are always date strings for consistency.

Other changes:

- When updating an entity, only the fields that have changed are sent to the server. When updating a list of entities, unmodified entities are not sent to the server at all. This dramatically reduces the payload size and should speed things up.
- When fetching entities, entity links are now typed as well. E.g. `project['rebateList']` contains a list of `Rebate` instead of `dict`.
- Added `date_from_timestamp()` and `timestamp_from_date()` functions to help with converting to/from dates and Projectal timestamps.
- Entity history now uses `desc` by default (index 0 is newest)
- Added `Project.tasks()` to list all task UUIDs within a project.

### 1.0.3

- Fix another case of automatic JWT refresh not working

### 1.0.2

- Entity instances can `save()` or `delete()` on themselves
- Fix broken `dict` methods (`get()` and `update()`) when called from Entity instances
- Fix automatic JWT refresh only working in some cases

### 1.0.1

- Added `list()` function for all entities
- Added search functions for all entities (match-, search, query)
- Added `Company.get_master_company()`
- Fixed adding template tasks

View Source

      1"""
      2A python client for the [Projectal API](https://projectal.com/docs/latest).
      3
      4## Getting started
      5
      6```
      7import projectal
      8import os
      9
     10# Supply your Projectal server URL and account credentials
     11projectal.api_base = 'https://yourcompany.projectal.com'
     12projectal.api_username = os.environ.get('PROJECTAL_USERNAME')
     13projectal.api_password = os.environ.get('PROJECTAL_PASSWORD')
     14
     15# Test communication with server
     16status = projectal.status()
     17
     18# Test account credentials
     19projectal.login()
     20details = projectal.auth_details()
     21```
     22
     23----
     24
     25## Changelog
     26
     27### 5.2.0
     28- Supported handling for new Projectal API rate limiting. When a request is rate limited,
     29  will pause for the required waiting period before retrying the request.
     30
     31### 5.1.0
     32- Added new Staff function: `projectal.Staff.create_contract()`.
     33  Allows creation of a new contract for the given Staff uuId. It's recommended to set a different
     34  Department, Position, Start/End Date or Pay Amount for new Contracts to differentiate them.
     35
     36  Parameters:
     37  - UUID: uuId of the Source Staff
     38  - payload: Optional payload to specify updated fields for the new Staff contract
     39  - end_current_contract (Default=False): Source Staff Contract will have its End Date set to Current Date.
     40    Source Staff Start Date must be before Current Date.
     41  - start_new_contract (Default=False): New Staff Contract will have its Start Date set to Current Date.
     42
     43- Allow fetching Staff entities with "CONTRACT" link. Will return list of all Contracts for each Staff.
     44
     45### 5.0.0
     46- Updated `projectal.login()` to use a basic Dict to store the login cookie instead of an instance
     47  of the RequestsCookieJar object from the Python Requests library. This fixes an issue where an
     48  indefinite loop can occur when the stored cookie is cleared unexpectedly after attempting to login
     49  again after a token expired.
     50- Update the stored cookie whenever a new cookie is returned by a successful request. This takes
     51  advantage of an updated Projectal implementation that refreshes the login cookie periodically.
     52  This should avoid the re-authentication procedure in most cases for scripts with long execution
     53  times.
     54
     55### 4.3.2
     56- Added CompanyType.Division enum, matching new system defaults.
     57
     58### 4.3.1
     59- Fixed typo in "Getting started" example.
     60- Fixed typo in 4.3.0 Changelog.
     61
     62### 4.3.0
     63
     64**Breaking changes:**
     65
     66- Added "PREDECESSOR_TASK" option when fetching projectal.Task with links
     67- Predecessor tasks will be returned as a list of tasks under the taskList attribute
     68- This attribute name is different to what you would see when using the REST API directly
     69  (planList). This change was necessary to allow for directly manipulating
     70  the list of links and then saving the task entity to commit any changes,
     71  since the REST API is expecting a different key for linking calls (taskList).
     72- Existing Predecessor Task linking methods were also updated to match the new linking functionality.
     73  They now work as a reverse linker, automatically inverting the link relationship between
     74  the task entities. This more closely matches what you would expect to see based on the Web UI.
     75  I.e. When previously calling `some_task.link_predecessor_task(another_task)`,
     76  some_task would be set as a predecessor for another_task instead of the other way around.
     77  Now another_task would be set as the predecessor for some_task.
     78
     79### 4.2.2
     80- `projectal.User.current_user_permissions()` fixed incorrect query.
     81- `projectal.Webhook.list()` default limit increased to 1000.
     82
     83### 4.2.1
     84- Added classes allowing for the management of dynamic enums. The user must have the "List Management"
     85  permission to update enums.
     86
     87  New classes:
     88  - `projectal.CompanyTypes`
     89  - `projectal.SkillLevels`
     90  - `projectal.StaffTypes`
     91  - `projectal.PriorityLevels`
     92  - `projectal.ComplexityLevels`
     93  - `projectal.CurrencyList`
     94
     95  The current enum can be retrieved with get(), and updated with set().
     96  For example, to return the current SkillLevels enum:
     97
     98  ```
     99  projectal.SkillLevels.get()
    100  ```
    101
    102  For each enum the entire list of key value pairs must be provided when calling set(),
    103  any existing values that are omitted from the dictionary will be removed,
    104  and any additional values will be added.
    105
    106  To update the SkillLevels enum with a new value:
    107
    108  ```
    109  new_value_added = {
    110      "Senior": 10,
    111      "Mid": 20,
    112      "Junior": 30,
    113      # new SkillLevel value "Beginner"
    114      "Beginner": 40,
    115  }
    116  projectal.SkillLevels.set(new_value_added)
    117  ```
    118
    119  To change the name of a value, set a new key name for the original value:
    120
    121  ```
    122  updated_value_name = {
    123      # changing "Senior" SkillLevel to "Expert"
    124      "Expert": 10,
    125      "Mid": 20,
    126      "Junior": 30,
    127  }
    128  projectal.SkillLevels.set(updated_value_name)
    129  ```
    130
    131  To remove an existing value, call set on a dictionary with that value removed:
    132
    133  ```
    134  value_removed = {
    135      "Senior": 10,
    136      "Mid": 20,
    137      # "Junior" SkillLevel removed
    138  }
    139  projectal.SkillLevels.set(new_value_added)
    140  ```
    141
    142  Updating the CurrencyList works differently to the other enums, since the
    143  names of values must match the alphabetic currency code and the value must
    144  match the numeric currency code.
    145  This will cause an exception if you try to change the name for any values.
    146
    147  Adding a new currency:
    148
    149  ```
    150  new_currency_added = {
    151    "AED": 784
    152    ...
    153    # rest of the existing currencies
    154    ...
    155    # new currency to add with the alphabetic and numeric code
    156    "ZWL": 932,
    157  }
    158  projectal.CurrencyList.set(new_currency_added)
    159  ```
    160
    161  Removing an existing currency requires you to provide the numeric code for
    162  the currency as a negative value.
    163
    164  ```
    165  currency_removed = {
    166    # this currency will be removed
    167    "AED": -784
    168    ...
    169    # rest of the existing currencies
    170    ...
    171  }
    172  projectal.CurrencyList.set(currency_removed)
    173  ```
    174
    175### 4.2.0
    176Version 4.2.0 accompanies the release of Projectal 4.1.0
    177
    178- Minimum Projectal version is now 4.1.0.
    179
    180- Changed order of applying link types when an entity is initialized,
    181prevents a type error with reverse linking in certain situations.
    182
    183- `projectal.Task.reset_duration()` now supports adjustments with multi day calendar exceptions.
    184
    185- `projectal.Task.reset_duration()` location working days override base exceptions.
    186
    187- `projectal.TaskTemplate.list()` fixed incorrect query when using inherited method.
    188
    189### 4.1.0
    190- DateLimit.Max enum value changed from "9999-12-31" to "3000-01-01". This reflects changes to the Projectal
    191backend that defines this as the maximum allowable date value. The front end typically considers this value as
    192equivalent with having no end date.
    193
    194- Updated requirements.txt version for requests package
    195
    196- Minimum Projectal version is now 4.0.40
    197
    198### 4.0.3
    199- When a dict object is passed to the update class method, it will be converted to the corresponding Entity type.
    200  Allows for proper handling of keys that require being treated as links.
    201
    202### 4.0.2
    203- Booking entity is now fetched with project field and either staff or resource field.
    204
    205- Added missing link methods for 'Booking' entity (Note, File)
    206
    207- Added missing link methods for 'Activity' entity (Booking, Note, File, Rebate)
    208
    209- Reduced maximum number of link methods to 100 for a single batch request to prevent timeouts
    210under heavy load.
    211
    212### 4.0.1
    213- Minimum Projectal version is now 4.0.0.
    214
    215### 4.0.0
    216
    217Version 4.0.0 accompanies the release of Projectal 4.0.
    218
    219- Added the `Activity` entity, new in Projectal 4.0.
    220
    221- Added the `Booking` entity, new in Projectal 4.0.
    222
    223### 3.1.1
    224- Link requests generated by 'projectal.Entity.create()' and 'projectal.Entity.update()' are now
    225  executed in batches. This is enabled by default with the 'batch_linking=True' parameter and can
    226  be disabled to execute each link request individually. It is recommended to leave this parameter
    227  enabled as this can greatly reduce the number of network requests.
    228
    229### 3.1.0
    230- Minimum Projectal version is now 3.1.5.
    231
    232- Added `projectal.Webhook.list_events()`. See API doc for details on how to use.
    233
    234- Added `deleted_at` parameter to `projectal.Entity.get()`. This value should be a UTC timestamp
    235  from a webhook delete event.
    236
    237- Added `projectal.ldap_sync()` to initiate a user sync with the LDAP/AD service configured in
    238  the Projectal server settings.
    239
    240- Enhanced output of `projectal.Entity.changes()` function when reporting link changes.
    241  It no longer dumps the entire before-and-after list with the full content of each linked entity.
    242  Now reports three lists: `added`, `updated`, `removed`. Entities within the `updated` list
    243  follow the same `old` vs `new` dictionary model for the data attributes within them. E.g:
    244
    245    ```
    246    resourceList: [
    247        'added': [],
    248        'updated': [
    249            {'uuId': '14eb4c31-0f92-49d1-8b4d-507ab939003e', 'resourceLink': {'utilization': {'old': 0.1, 'new': 0.9}}},
    250        ],
    251        'removed': []
    252    ]
    253    ```
    254  This should result in slimmer logs that are much easier to understand as the changes are
    255  clearly indicated.
    256
    257### 3.0.2
    258- Added `projectal.Entity.get_link_definitions()`. Exposes entity link definition dictionary.
    259  Consumers can inspect which links an Entity knows about and their internal settings.
    260  Link definitions that appear here are the links valid for `links=[]` parameters.
    261
    262### 3.0.1
    263- Fixed fetching project with links=['task'] not being available.
    264
    265- Improved Permission.list(). Now returns a dict with the permission name as
    266  key with Permission objects as the value (instead of list of uuIds).
    267
    268- Added a way to use the aliasing feature of the API (new in Projectal 3.0).
    269Set `projectal.api_alias = 'uuid'` to the UUID of a User object and all
    270requests made will be done as that user. Restore this value to None to resume
    271normal operation. (Some rules and limitations apply. See API for more details.)
    272
    273- Added complete support for the Tags entity (including linkers).
    274
    275### 3.0
    276
    277Version 3.0 accompanies the release of Projectal 3.0.
    278
    279**Breaking changes**:
    280
    281- The `links` parameter on `Entity` functions now consumes a list of entity
    282  names instead of a comma-separated string. For example:
    283
    284    ```
    285    # Before:
    286    projectal.Staff.get('<uuid>', links='skill,location')  # No longer valid
    287    # Now:
    288    projectal.Staff.get('<uuid>', links=['skill', 'location'])
    289    ```
    290
    291- The `projectal.enums.SkillLevel` enum has had all values renamed to match the new values
    292  used in Projectal (Junior, Mid, Senior). This includes the properties on
    293  Skill entities indicating work time for auto-scheduling (now `juniorLevel`,
    294  `midLevel`, `seniorLevel`).
    295
    296**Other changes**:
    297
    298- Working with entity links has changed in this release. The previous methods
    299  are still available and continue to work as before, but there is no need
    300  to interact with the `projectal.linkers` methods yourself anymore.
    301
    302  You can now modify the list of links within an entity and save the entity
    303  directly. The library will automatically determine how the links have been
    304  modified and issue the correct linker methods on your behalf. E.g.,
    305  you can now do:
    306
    307    ```
    308    staff = projectal.Staff.get('<uuid>', links=['skill'])
    309    staff['firstName'] = "New name"  # Field update
    310    staff['skillList'] = [skill1, skill2, skill3]  # Link update
    311    staff.save()  # Both changes are saved
    312
    313    task = projectal.Task.get('<uuid>', links=['stage'])
    314    task['stage'] = stage1  # Uses a single object instead of list
    315    task.save()
    316    ```
    317
    318  See `examples/linking.py` for a more complete demonstration of linking
    319  capabilities and limitations.
    320
    321- Linkers (`projectal.linkers`) can now be given a list of Entities (of one
    322 type) to link/unlink/relink in bulk. E.g:
    323    ```
    324    staff.unlink_skill(skill1)  # Before
    325    staff.unlink_skill([skill1, skill2, skill3])  # This works now too
    326    ```
    327
    328- Linkers now strip the payload to only the required fields instead of passing
    329  on the entire Entity object. This cuts down on network traffic significantly.
    330
    331- Linkers now also work in reverse. The Projectal server currently only supports
    332  linking entities in one direction (e.g., Company to Staff), which often means
    333  writing something like:
    334    ```
    335    staff.link_location(location)
    336    company.link_staff(staff)
    337    ```
    338  The change in direction is not very intuitive and would require you to constantly
    339  verify which direction is the one available to you in the documentation.
    340
    341  Reverse linkers hide this from you and figure out the direction of the relationship
    342  for you behind the scenes. So now this is possible, even though the API doesn't
    343  strictly support it:
    344    ```
    345    staff.link_location(location)
    346    staff.link_company(company)
    347    ```
    348    Caveat: the documentation for Staff will not list Company links. You will still
    349    have to look up the Company documentation for the link description.
    350
    351- Requesting entity links with the `links=` parameter will now always ensure the
    352  link field (e.g., `taskList`) exists in the result, even if there are no links.
    353  The server may not always return a value, but we can use a default value ([] for
    354  lists, None for dicts).
    355
    356- Added a `Permission` entity to correctly type Permissions in responses.
    357
    358- Added a `Tag` entity, new in Projectal 3.0.
    359
    360- Added `links` parameter to `Company.get_primary_company()`
    361
    362- `Department.tree()`: now consumes a `holder` Entity object instead
    363  of a uuId.
    364
    365- `Department.tree()`: added `generic_staff` parameter, new in
    366  Projectal 3.0.
    367
    368- Don't break on trailing slash in Projectal URL
    369
    370- When creating tasks, populate the `projectRef` and `parent` fields in the
    371  returned Task object.
    372
    373- Added convenience functions for matching on fields where you only want
    374  one result (e.g match_one()) which return the first match found.
    375
    376- Update the entity `history()` method for Projectal 3.0. Some new parameters
    377  allow you to restrict the history to a particular range or to get only the
    378  changes for a webhook timestamp.
    379
    380- Entity objects can call `.history()` on themselves.
    381
    382- The library now keeps a reference to the User account that is currently logged
    383  in and using the API: `projectal.api_auth_details`.
    384
    385**Known issues**:
    386- You cannot save changes to Notes or Calendars via their holding entity. You
    387  must save the changes on the Note or Calendar directly. To illustrate:
    388  ```
    389  staff = projectal.Staff.get(<uuid>, links=['calendar'])
    390  calendar = staff['calendarList'][0]
    391  calendar['name'] = 'Calendar 2'
    392
    393  # Cannot do this - will not pick up the changes
    394  staff.save()
    395
    396  # You must do this for now
    397  calendar.save()
    398  ```
    399  This will be resolved in a future release.
    400
    401- When creating Notes, the `created` and `modified` values may differ by
    402  1ms in the object you have a reference to compared to what is actually
    403  stored in the database.
    404
    405- Duration calculation is not precise yet (mentioned in 2.1.0)
    406
    407### 2.1.0
    408**Breaking changes**:
    409- Getting location calendar is now done on an instance instead of class. So
    410  `projectal.Location.calendar(uuid)` is now simply `location.calendar()`
    411- The `CompanyType.Master` enum has been replaced with `CompanyType.Primary`.
    412  This was a leftover reference to the Master Company which was renamed in
    413  Projectal several versions ago.
    414
    415**Other changes**:
    416- Date conversion functions return None when given None or empty string
    417- Added `Task.reset_duration()` as a basic duration calculator for tasks.
    418  This is a work-in-progress and will be gradually improved. The duration
    419  calculator takes into consideration the location to remove non-work
    420  days from the estimate of working duration. It currently does not work
    421  for the time component or `isWorking=True` exceptions.
    422- Change detection in `Entity.changes()` now excludes cases where the
    423  server has no value and the new value is None. Saving this change has
    424  no effect and would always detect a change until a non-None value is
    425  set, which is noisy and generates more network activity.
    426
    427### 2.0.3
    428- Better support for calendars.
    429  - Distinguish between calendar containers ("Calendar") and the
    430    calendar items within them ("CalendarItem").
    431  - Allow CalendarItems to be saved directly. E.G item.save()
    432- Fix 'holder' parameter in contact/staff/location/task_template not
    433  permitting object type. Now consumes uuId or object to match rest of
    434  the library.
    435- `Entity.changes()` has been extended with an `old=True` flag. When
    436  this flag is true, the set of changes will now return both the original
    437  and the new values. E.g.
    438```
    439task.changes()
    440# {'name': 'current'}
    441task.changes(old=True)
    442# {'name': {'old': 'original', 'new': 'current'}}
    443```
    444- Fixed entity link cache causing errors when deleting a link from an entity
    445  which has not been fetched with links (deleting from empty list).
    446
    447### 2.0.2
    448- Fixed updating Webhook entities
    449
    450### 2.0.1
    451- Fixed application ID not being used correctly.
    452
    453### 2.0.0
    454- Version 2.0 accompanies the release of Projectal 2.0. There are no major changes
    455  since the previous release.
    456- Expose `Entity.changes()` function. It returns a list of fields on an entity that
    457  have changed since fetching it. These are the changes that will be sent over to the
    458  server when an update request is made.
    459- Added missing 'packaging' dependency to requirements.
    460
    461### 1.2.0
    462
    463**Breaking changes**:
    464
    465- Renamed `request_timestamp` to `response_timestamp` to better reflect its purpose.
    466- Automatic timestamp conversion into dates (introduced in `1.1.0`) has been reverted.
    467  All date fields returned from the server remain as UTC timestamps.
    468
    469  The reason is that date fields on tasks contain a time component and converting them
    470  into date strings was erasing the time, resulting in a value that does not match
    471  the database.
    472
    473  Note: the server supports setting date fields using a date string like `2022-04-05`.
    474  You may use this if you prefer but the server will always return a timestamp.
    475
    476  Note: we provide utility functions for easily converting dates from/to
    477  timestamps expected by the Projectal server. See:
    478  `projectal.date_from_timestamp()`,`projectal.timestamp_from_date()`, and
    479  `projectal.timestamp_from_datetime()`.
    480
    481**Other changes**:
    482- Implement request chunking - for methods that consume a list of entities, we now
    483  automatically batch them up into multiple requests to prevent timeouts on really
    484  large request. Values are configurable through
    485  `projectal.chunk_size_read` and `projectal.chunk_size_write`.
    486  Default values: Read: 1000 items. Write: 200 items.
    487- Added profile get/set functions on entities for easier use. Now you only need to supply
    488  the key and the data. E.g:
    489
    490```
    491key = 'hr_connector'
    492data = {'staff_source': 'company_z'}
    493task.profile_set(key, data)
    494```
    495
    496- Entity link methods now automatically update the entity's cached list of links. E.g:
    497  a task fetched with staff links will have `task['staffList'] = [Staff1,Staff2]`.
    498  Before, doing a `task.link_staff(staff)` did not modify the list to reflect the
    499  addition. Now, it will turn into `[Staff1,Staff2,Staff3]`. The same applies for update
    500  and delete.
    501
    502  This allows you to modify links and continue working with that object without having
    503  to fetch it again to obtain the most recent link data. Be aware that if you acquire
    504  the object without requesting the link data as well
    505  (e.g: `projectal.Task.get(id, links='STAFF')`),
    506  these lists will not accurately reflect what's in the database, only the changes made
    507  while the object is held.
    508
    509- Support new `applicationId` property on login. Set with: `projectal.api_application_id`.
    510  The application ID is sent back to you in webhooks so you know which application was
    511  the source of the event (and you can choose to filter them accordingly).
    512- Added `Entity.set_readonly()` to allow setting values on entities that will not
    513  be sent over to the server when updating/saving the entity.
    514
    515  The main use case for this is to populate cached entities which you have just created
    516  with values you already know about. This is mainly a workaround for the limitation of
    517  the server not sending the full object back after creating it, resulting in the client
    518  needing to fetch the object in full again if it needs some of the fields set by the
    519  server after creation.
    520
    521  Additionally, some read-only fields will generate an error on the server if
    522  included in the update request. This method lets you set these values on newly
    523  created objects without triggering this error.
    524
    525  A common example is setting the `projectRef` of a task you just created.
    526
    527
    528### 1.1.1
    529- Add support for 'profiles' API. Profiles are a type of key-value storage that target
    530  any entity. Not currently documented.
    531- Fix handling error message parsing in ProjectalException for batch create operation
    532- Add `Task.update_order()` to set task order
    533- Return empty list when GETing empty list instead of failing (no request to server)
    534- Expose the timestamp returned by requests that modify the database. Use
    535  `projectal.request_timestamp` to get the value of the most recent request (None
    536  if no timestamp in response)
    537
    538### 1.1.0
    539- Minimum Projectal version is now 1.9.4.
    540
    541**Breaking changes**:
    542- Entity `list()` now returns a list of UUIDs instead of full objects. You may provide
    543  an `expand` parameter to restore the previous behavior: `Entity.list(expand=True)`.
    544  This change is made for performance reasons where you may have thousands of tasks
    545  and getting them all may time out. For those cases, we suggest writing a query to filter
    546  down to only the tasks and fields you need.
    547- `Company.get_master_company()` has been renamed to `Company.get_primary_company()`
    548  to match the server.
    549- The following date fields are converted into date strings upon fetch:
    550  `startTime`, `closeTime`, `scheduleStart`, `scheduleFinish`.
    551  These fields are added or updated using date strings (like `2022-03-02`), but the
    552  server returns timestamps (e.g: 1646006400000) upon fetch, which is confusing. This
    553  change ensures they are always date strings for consistency.
    554
    555**Other changes**:
    556- When updating an entity, only the fields that have changed are sent to the server. When
    557  updating a list of entities, unmodified entities are not sent to the server at all. This
    558  dramatically reduces the payload size and should speed things up.
    559- When fetching entities, entity links are now typed as well. E.g. `project['rebateList']`
    560  contains a list of `Rebate` instead of `dict`.
    561- Added `date_from_timestamp()` and `timestamp_from_date()` functions to help with
    562  converting to/from dates and Projectal timestamps.
    563- Entity history now uses `desc` by default (index 0 is newest)
    564- Added `Project.tasks()` to list all task UUIDs within a project.
    565
    566### 1.0.3
    567- Fix another case of automatic JWT refresh not working
    568
    569### 1.0.2
    570- Entity instances can `save()` or `delete()` on themselves
    571- Fix broken `dict` methods (`get()` and `update()`) when called from Entity instances
    572- Fix automatic JWT refresh only working in some cases
    573
    574### 1.0.1
    575- Added `list()` function for all entities
    576- Added search functions for all entities (match-, search, query)
    577- Added `Company.get_master_company()`
    578- Fixed adding template tasks
    579
    580"""
    581import logging
    582import os
    583
    584from projectal.entities import *
    585from projectal.dynamic_enums import *
    586from .api import *
    587from . import profile
    588
    589api_base = os.getenv("PROJECTAL_URL")
    590api_username = os.getenv("PROJECTAL_USERNAME")
    591api_password = os.getenv("PROJECTAL_PASSWORD")
    592api_application_id = None
    593api_auth_details = None
    594api_alias = None
    595cookies = None
    596chunk_size_read = 1000
    597chunk_size_write = 200
    598
    599# Records the timestamp generated by the last request (database
    600# event time). These are reported on add or updates; if there is
    601# no timestamp in the response, this is set to None.
    602response_timestamp = None
    603
    604
    605# The minimum version number of the Projectal instance that this
    606# API client targets. Lower versions are not supported and will
    607# raise an exception.
    608MIN_PROJECTAL_VERSION = "5.3.0"
    609
    610__verify = True
    611
    612logging.getLogger("projectal-api-client").addHandler(logging.NullHandler())

api_base = None
api_username = None
api_password = None
api_application_id = None
api_auth_details = None
api_alias = None
cookies = None
chunk_size_read = 1000
chunk_size_write = 200
response_timestamp = None
MIN_PROJECTAL_VERSION = '5.3.0'